#!/usr/bin/env python2
#
# Piotr Nyczyk <Piotr.Nyczyk@cern.ch>, Judit Novak <Judit.Novak@cern.ch>
# Last modification: Fri Jul 29 11:31:40 CEST 2005
#

# Can be started as standalone program or loaded as a module

import rgma
import time
import re
import pickle


CONTINUOUS=rgma.QueryType.CONTINUOUS
LATEST=rgma.QueryType.LATEST
HISTORY=rgma.QueryType.HISTORY



class BasicQuery (object):
    """Abstraction layer for R-GMA queries that makes query
    behave like a python iterator returning subsequent tuples"""


    def deserialize(cls, file):
    
        dq=BasicQuery.__new__(BasicQuery)
        datadict = pickle.load(open(file, 'r'))
        dq.cons = rgma.reconnect(rgma.Consumer, rgma.ResourceEndpoint(datadict['URL'], datadict['ID']))
        dq.buffer = datadict['buffer']
        dq.sql = datadict['sql']
        return dq

    deserialize=classmethod(deserialize)


    def __init__(self,sql,type=LATEST,timeout=300,termination=300):

        terminationInterval=rgma.TimeInterval(termination)
        timeoutInterval=rgma.TimeInterval(timeout)
        self.cons=rgma.Consumer(terminationInterval,
                           rgma.QueryProperties(type,timeoutInterval),
                           sql)
        self.sql = sql
        self.cons.start(timeoutInterval)
        self.buffer=[]


    def __close__(self):
        self.cons.close()        


    def __iter__(self):
        return self


    def next(self):
        "Return next value or stop iteration"
        # make sure buffer is not empty unless the query finished
        while (not self.buffer) and self.cons.isExecuting():
            self.buffer=self.cons.pop()
            time.sleep(0.1)

        # get the remaining data 
        if not self.buffer:
            self.buffer.extend(self.cons.pop())
        
        if self.buffer:
            return self.buffer.pop(0) # next value
        else:
            raise StopIteration       # no more data - stop


    def count(self):
        return self.cons.count() + len(self.buffer)


    def count_existing(self):
        
        query = re.sub("SELECT.+ FROM ", "SELECT COUNT(*) FROM ", self.sql )   
        
        cons = rgma.Consumer(rgma.TimeInterval(10),
                           rgma.QueryProperties(rgma.QueryType.LATEST,rgma.TimeInterval(100)),
                           query)
        cons.start(rgma.TimeInterval(100))
        
        while cons.isExecuting():
            time.sleep(1)

        tuple = cons.pop()
        cons.close()
        if tuple:
            return tuple[0]
        else:
            return None

        
    def still_exec(self):
        return self.cons.isExecuting()



    def serialize(self, file):

        datadict={}
        datadict['ID'] = self.cons.getEndpoint().getResourceId()
        datadict['URL'] = self.cons.getEndpoint().getUrl()
        datadict['buffer'] = self.buffer
        datadict['sql'] = self.sql
        pickle.dump(datadict, open(file, 'w')) 




class DetailedQuery (BasicQuery):

    def deserialize(cls, file):
    
        dq=DetailedQuery.__new__(DetailedQuery)
        datadict = pickle.load(open(file, 'r'))
        dq.cons = rgma.reconnect(rgma.Consumer, rgma.ResourceEndpoint(datadict['URL'], datadict['ID']))
        dq.buffer = datadict['buffer']
        dq.sql = datadict['sql']
        dq.tables = datadict['tables']
        dq.cols = datadict['cols']
        dq.filters = datadict['filters']
        return dq

    deserialize=classmethod(deserialize)

    def __init__(self, tables,  cols=['*'], filters=[] ,type='LATEST',timeout=300,termination=300):

        # Checking parameters and constructing query
        for col in cols:
            if not re.compile( "(\s*(\*|\w+)\s*,?)+" ).match( col ):
                raise ValueError( col + ": attribute error: '[<col1>, <col2>, ...]'")

        for filter in filters:
            if not re.compile("\(?\s*\w+\s*=\s*['\w%]+\s*\)?").match( filter ):
                raise ValueError( filter + ": attribute error: '[<col1>=<val1>, <col2>=<val2>, ...]'")
       
        
        if type.lower() == 'history':
            type = rgma.QueryType.HISTORY
        elif type.lower() == 'continuous': 
            type = rgma.QueryType.CONTINUOUS
        elif type.lower() == 'latest': 
            type = rgma.QueryType.LATEST
        else:   
            raise ValueError( "Invalid query type '" + type + "'" )

        # Setting private parameters
        self.tables = tables
        self.cols = cols
        self.filters = filters
        
        # Forming query
        tablestr = ", ".join(tables)
        colsstr = ", ".join(cols)
        sql = "SELECT " + colsstr + " FROM " + tablestr 
        if filters:
            filterstr = " AND ".join(filters)
            sql = sql + " WHERE " + filterstr

        BasicQuery.__init__(self,sql,type,timeout,termination)


    def serialize(self, file):

        datadict={}
        datadict['ID'] = self.cons.getEndpoint().getResourceId()
        datadict['URL'] = self.cons.getEndpoint().getUrl()
        datadict['buffer'] = self.buffer
        datadict['sql'] = self.sql
        datadict['tables'] = self.tables
        datadict['cols'] = self.cols
        datadict['filters'] = self.filters
        pickle.dump(datadict, open(file, 'w')) 



class AndOrQuery (DetailedQuery):


    def __init__(self, tables,  cols=['*'], filters=[] ,type='LATEST',timeout=300,termination=300):

        filters2=[]
        if filters:
            for filter in filters:
                if isinstance(filter, list):
                    filterstr = " OR ".join(filter)
                    filterstr = '(' + filterstr + ')'
                    filters2.append(filterstr)
                else:
                    filters2.append(filter)
                    
        DetailedQuery.__init__(self, tables, cols, filters2, type, timeout, termination)

        
def timingTest():
    startTime=time.time()
    firstTime=startTime
    n=0
    for tuple in BasicQuery("select * from userTable",HISTORY):
        t=time.time()
        if not n:
            firstTime=t
            print "Got first tuple after %g seconds"%(firstTime-startTime)
        n+=1
        if not n%1000:
            print "already got %6d tuples in %g seconds "%(n,t-startTime)

    endTime=time.time()
    print
    print "Query finished"
    print "First tuple arrived after %g seconds"%(firstTime-startTime)
    print "Got %d tuples in %g seconds"%(n,endTime-startTime)

# main - run only if started as program and not if loaded as a module

if __name__=='__main__':
    timingTest()
